@@ -5,10 +5,12 @@ use strict;
 use warnings;
 use Log::ger;
 
+use App::orgadb::Common;
+
 our $AUTHORITY = 'cpan:PERLANCAR'; # AUTHORITY
-our $DATE = '2022-06-24'; # DATE
+our $DATE = '2022-07-03'; # DATE
 our $DIST = 'App-orgadb'; # DIST
-our $VERSION = '0.002'; # VERSION
+our $VERSION = '0.003'; # VERSION
 
 our %SPEC;
 
@@ -17,42 +19,6 @@ $SPEC{':package'} = {
     summary => 'An opinionated Org addressbook tool',
 };
 
-our %argspecs_common = (
-    files => {
-        summary => 'Path to addressbook files',
-        'summary.alt.plurality.singular' => 'Path to addressbook file',
-        'x.name.is_plural' => 1,
-        'x.name.singular' => 'file',
-        schema => ['array*', of=>'filename*', min_len=>1],
-        cmdline_aliases=>{f=>{}},
-        tags => ['category:input'],
-    },
-);
-
-our %argspec_category = (
-    entry => {
-        summary => 'Find entry by string or regex search against the category title',
-        schema => 'str_or_re*',
-        cmdline_aliases=>{c=>{}},
-    },
-);
-
-our %argspecopt0_entry = (
-    entry => {
-        summary => 'Find entry by string or regex search against its title',
-        schema => 'str_or_re*',
-        pos => 0,
-    },
-);
-
-our %argspecopt1_field = (
-    field => {
-        summary => 'Find field by string or regex search',
-        schema => 'str_or_re*',
-        pos => 1,
-    },
-);
-
 sub _highlight {
     my ($clrtheme_obj, $re, $text) = @_;
     return $text unless $clrtheme_obj && $re;
@@ -63,77 +29,19 @@ sub _highlight {
     $text;
 }
 
-$SPEC{select_addressbook_entries} = {
-    v => 1.1,
-    summary => 'Select Org document elements using CSel (CSS-selector-like) syntax',
-    args => {
-        %argspecs_common,
-        %argspecopt0_entry,
-        %argspecopt1_field,
-        category => {
-            schema => 'str_or_re*',
-            cmdline_aliases=>{c=>{}},
-        },
-        hide_category => {
-            summary => 'Do not show category',
-            schema => 'true*',
-            cmdline_aliases => {C=>{}},
-        },
-        hide_entry => {
-            summary => 'Do not entry headline',
-            schema => 'true*',
-            cmdline_aliases => {E=>{}},
-        },
-        color => {
-            summary => 'Whether to use color',
-            schema => ['str*', in=>[qw/auto always never/]],
-            default => 'auto',
-        },
-        color_theme => {
-            schema => 'perl::colortheme::modname_with_optional_args*',
-        },
-        detail => {
-            schema => 'bool*',
-            cmdline_aliases => {l=>{}},
-        },
-    },
-    'x.envs' => {
-        'ORGADB_COLOR_THEME' => {
-            summary => 'Set default color theme',
-            schema => 'perl::colortheme::modname_with_optional_args*',
-            description => <<'_',
-
-Color theme is Perl module name under the `ColorTheme::Search::` namespace,
-without the namespace prefix. The default is `Light`. You can set color theme
-using the `--color-theme` command-line option as well as this environment
-variable.
-
-_
-        },
-    },
-};
-sub select_addressbook_entries {
+# this is like select_addressbook_entries(), but selects from object trees
+# instead of from an Org file.
+sub _select_addressbook_entries_single {
     my %args = @_;
 
-    my @trees;
-  PARSE_FILES: {
-        require Org::Parser;
-        my $parser = Org::Parser->new;
+    #print "$_ => $args{$_}\n" for sort keys %args;
 
-        for my $file (@{ $args{files} }) {
-            my $doc;
-            if ($file eq '-') {
-                binmode STDIN, ":encoding(utf8)";
-                $doc = $parser->parse(join "", <>);
-            } else {
-                local $ENV{PERL_ORG_PARSER_CACHE} = $ENV{PERL_ORG_PARSER_CACHE} // 1;
-                $doc = $parser->parse_file($file);
-            }
-            push @trees, $doc;
-        } # for file
-    } # PARSe_FILES
+    my $trees = $args{_trees};
+    my $tree_filenames = $args{_tree_filenames};
 
-    my @entries;
+    my $res = [200, "OK", ""];
+
+    my @matching_entries;
     my ($re_category, $re_entry, $re_field);
   FIND_ENTRIES: {
         require Data::CSel;
@@ -167,14 +75,14 @@ sub select_addressbook_entries {
         log_trace "CSel expression: <$expr>";
         #log_trace "Number of trees: %d", scalar(@trees);
 
-        for my $tree (@trees) {
+        for my $tree (@$trees) {
             my @nodes = Data::CSel::csel({
                 class_prefixes => ["Org::Element"],
             }, $expr, $tree);
-            push @entries, @nodes;
+            push @matching_entries, @nodes;
         }
     } # FIND_ENTRIES
-    log_trace "Number of matching entries: %d", scalar(@entries);
+    log_trace "Number of matching entries: %d", scalar(@matching_entries);
 
   DISPLAY_ENTRIES: {
         my ($clrtheme, $clrtheme_obj);
@@ -192,64 +100,162 @@ sub select_addressbook_entries {
                 {ns_prefixes=>['ColorTheme::Search','ColorTheme','']}, $clrtheme);
         };
 
-        my ($re_field, $expr_field);
+        my ($expr_field, @re_field);
       ENTRY:
-        for my $entry (@entries) {
+        for my $entry (@matching_entries) {
 
-            my @fields;
-            if (defined $args{field}) {
+            my @matching_fields;
+            if (defined($args{fields}) && @{ $args{fields} }) {
                 unless (defined $expr_field) {
                     $expr_field = '';
-                    $expr_field .= 'ListItem[desc_term.text';
-                    if (ref $args{field} eq 'Regexp') {
-                        $re_field = $args{field};
-                    } else {
-                        $re_field = quotemeta($args{field});
-                        $re_field = qr/$re_field/;
+                    for my $field_term (@{ $args{fields} }) {
+                        $expr_field .= ' ' if $expr_field;
+                        $expr_field .= 'ListItem[desc_term.text';
+                        my $re_field;
+                        if (ref $field_term eq 'Regexp') {
+                            $re_field = $field_term;
+                        } else {
+                            $re_field = quotemeta($field_term);
+                            $re_field = qr/$re_field/;
+                        }
+                        $expr_field .= " =~ " . Data::Dmp::dmp($re_field) . "]";
+                        push @re_field, $re_field;
                     }
-                    $expr_field .= " =~ " . Data::Dmp::dmp($re_field) . "]";
                 }
 
-                @fields = Data::CSel::csel({
+                @matching_fields = Data::CSel::csel({
                     class_prefixes => ["Org::Element"],
                 }, $expr_field, $entry);
 
-                next ENTRY unless @fields;
+                next ENTRY unless @matching_fields;
             }
 
-            unless ($args{detail} && $args{hide_entry}) {
+            unless ($args{hide_entry}) {
+                $res->[2] .= "** ";
                 unless ($args{hide_category}) {
-                    print _highlight(
+                    $res->[2] .= _highlight(
                         $clrtheme_obj,
                         $re_category,
                         $entry->parent->title->text) . "/";
                 }
-                print _highlight(
+                $res->[2] .= _highlight(
                     $clrtheme_obj,
                     $re_entry,
                     $entry->title->text,
                 );
-                print "\n";
+                $res->[2] .= "\n";
             }
 
-            if ($args{detail} && !defined($args{field})) {
-                print $entry->children_as_string;
-            } elsif (@fields) {
-                for my $field (@fields) {
-                    my $str = _highlight(
-                        $clrtheme_obj,
-                        $re_field,
-                        $field->desc_term->text,
-                    ) . " ::" . $field->children_as_string;
-                    $str =~ s/^/  /gm;
-                    print $str;
+            my $re_field;
+            $re_field = join "|", @re_field if @re_field;
+            if ($args{detail}) {
+                my $str = $entry->children_as_string;
+                $str = _highlight(
+                    $clrtheme_obj,
+                    $re_field,
+                    $str) if defined $re_field;
+                $res->[2] .= $str;
+            } elsif (@matching_fields) {
+                for my $field (@matching_fields) {
+
+                    unless ($args{hide_field_name}) {
+                        my $field_name = '';
+                        $field_name = _highlight(
+                            $clrtheme_obj,
+                            $re_field,
+                            $field->bullet . ' ' . $field->desc_term->text,
+                        ) . " ::";
+                        $res->[2] .= $field_name;
+                    }
+
+                    my $field_value = $field->children_as_string;
+                    $field_value =~ s/\A\s+//s if $args{hide_field_name};
+                    $res->[2] .= $field_value;
                 }
             }
         }
     }
 
+    $res;
+}
+
+sub _select_addressbook_entries_shell {
+    my %args = @_;
+
+    require App::orgadb::Shell;
+    my $shell = App::orgadb::Shell->new(
+        orgadb_args => \%args,
+    );
+
+    $shell->cmdloop;
     [200];
 }
+
+$SPEC{select_addressbook_entries} = {
+    v => 1.1,
+    summary => 'Select Org addressbook entries/fields/subfields',
+    args => {
+        %App::orgadb::Common::argspecs_common,
+        %App::orgadb::Common::argspecs_select,
+    },
+    'x.envs' => {
+        'ORGADB_COLOR_THEME' => {
+            summary => 'Set default color theme',
+            schema => 'perl::colortheme::modname_with_optional_args*',
+            description => <<'_',
+
+Color theme is Perl module name under the `ColorTheme::Search::` namespace,
+without the namespace prefix. The default is `Light`. You can set color theme
+using the `--color-theme` command-line option as well as this environment
+variable.
+
+_
+        },
+    },
+};
+sub select_addressbook_entries {
+    my %args = @_;
+
+    my $code_parse_files = sub {
+        my @filenames = @_;
+
+        my @trees;
+        my @tree_filenames;
+
+        require Org::Parser;
+        my $parser = Org::Parser->new;
+
+        for my $filename (@filenames) {
+            my $doc;
+            if ($filename eq '-') {
+                binmode STDIN, ":encoding(utf8)";
+                $doc = $parser->parse(join "", <>);
+            } else {
+                local $ENV{PERL_ORG_PARSER_CACHE} = $ENV{PERL_ORG_PARSER_CACHE} // 1;
+                $doc = $parser->parse_file($filename);
+            }
+            push @trees, $doc;
+            push @tree_filenames, $filename;
+        } # for filename
+
+        return (\@trees, \@tree_filenames);
+    };
+
+    if ($args{shell}) {
+        _select_addressbook_entries_shell(
+            _code_parse_files => $code_parse_files,
+            %args,
+        );
+    } else {
+        my ($trees, $tree_filenames) = $code_parse_files->(@{ $args{files} });
+        _select_addressbook_entries_single(
+            %args,
+            _trees => $trees,
+            _tree_filenames => $tree_filenames,
+        );
+    }
+}
+
 1;
 # ABSTRACT: An opinionated Org addressbook tool
 
@@ -265,7 +271,7 @@ App::orgadb - An opinionated Org addressbook tool
 
 =head1 VERSION
 
-This document describes version 0.002 of App::orgadb (from Perl distribution App-orgadb), released on 2022-06-24.
+This document describes version 0.003 of App::orgadb (from Perl distribution App-orgadb), released on 2022-07-03.
 
 =head1 SYNOPSIS
 
@@ -278,7 +284,7 @@ Usage:
 
  select_addressbook_entries(%args) -> [$status_code, $reason, $payload, \%result_meta]
 
-Select Org document elements using CSel (CSS-selector-like) syntax.
+Select Org addressbook entriesE<sol>fieldsE<sol>subfields.
 
 This function is not exported.
 
@@ -288,6 +294,8 @@ Arguments ('*' denotes required arguments):
 
 =item * B<category> => I<str_or_re>
 
+Find entry by string or regex search against the category title.
+
 =item * B<color> => I<str> (default: "auto")
 
 Whether to use color.
@@ -300,9 +308,9 @@ Whether to use color.
 
 Find entry by string or regex search against its title.
 
-=item * B<field> => I<str_or_re>
+=item * B<fields> => I<array[str_or_re]>
 
-Find field by string or regex search.
+Find (sub)fields by string or regex search.
 
 =item * B<files> => I<array[filename]>
 
@@ -314,7 +322,15 @@ Do not show category.
 
 =item * B<hide_entry> => I<true>
 
-Do not entry headline.
+Do not show entry headline.
+
+=item * B<hide_field_name> => I<true>
+
+Do not show field names, just show field values.
+
+=item * B<reload_files_on_change> => I<bool> (default: 1)
+
+=item * B<shell> => I<true>
 
 
 =back
